@@ -1,5 +1,5 @@
 /* md.c  -  message digest dispatcher
- * Copyright (C) 1998, 1999, 2002, 2003 Free Software Foundation, Inc.
+ * Copyright (C) 1998, 1999, 2002, 2003, 2006 Free Software Foundation, Inc.
  *
  * This file is part of Libgcrypt.
  *
@@ -56,6 +56,7 @@ static struct digest_table_entry
 #endif
 #if USE_SHA256
     { &_gcry_digest_spec_sha256, GCRY_MD_SHA256 },
+    { &_gcry_digest_spec_sha224, GCRY_MD_SHA224 },
 #endif
 #if USE_SHA512
     { &_gcry_digest_spec_sha512, GCRY_MD_SHA512 },
@@ -63,6 +64,9 @@ static struct digest_table_entry
 #endif
 #if USE_TIGER
     { &_gcry_digest_spec_tiger, GCRY_MD_TIGER },
+#endif
+#if USE_WHIRLPOOL
+    { &_gcry_digest_spec_whirlpool, GCRY_MD_WHIRLPOOL },
 #endif
     { NULL },
   };
@@ -97,6 +101,7 @@ struct gcry_md_context
   int finalized;
   GcryDigestEntry *list;
   byte *macpads;
+  int macpads_Bsize;             /* Blocksize as used for the HMAC pads. */
 };
 
 
@@ -125,14 +130,14 @@ static gcry_err_code_t md_open (gcry_md_hd_t *h, int algo,
 static gcry_err_code_t md_enable (gcry_md_hd_t hd, int algo);
 static gcry_err_code_t md_copy (gcry_md_hd_t a, gcry_md_hd_t *b);
 static void md_close (gcry_md_hd_t a);
-static void md_write (gcry_md_hd_t a, byte *inbuf, size_t inlen);
+static void md_write (gcry_md_hd_t a, const void *inbuf, size_t inlen);
 static void md_final(gcry_md_hd_t a);
 static byte *md_read( gcry_md_hd_t a, int algo );
 static int md_get_algo( gcry_md_hd_t a );
 static int md_digest_length( int algo );
 static const byte *md_asn_oid( int algo, size_t *asnlen, size_t *mdlen );
-static void md_start_debug( gcry_md_hd_t a, char *suffix );
-static void md_stop_debug( gcry_md_hd_t a );
+static void md_start_debug ( gcry_md_hd_t a, const char *suffix );
+static void md_stop_debug ( gcry_md_hd_t a );
 
 
 
@@ -298,7 +303,7 @@ gcry_md_map_name (const char *string)
   ret = search_oid (string, &algorithm, NULL);
   if (! ret)
     {
-      /* Not found, search for an acording diget name.  */
+      /* Not found, search a matching digest name.  */
       digest = gcry_md_lookup_name (string);
       if (digest)
 	{
@@ -426,11 +431,17 @@ md_open (gcry_md_hd_t *h, int algo, int secure, int hmac)
 
       if (hmac)
 	{
-	  ctx->macpads = gcry_malloc_secure (128);
+	  if ( (GCRY_MD_SHA384 == algo) || (GCRY_MD_SHA512 == algo) ) {
+	    ctx->macpads_Bsize = 128;
+	    ctx->macpads = gcry_malloc_secure (2*(ctx->macpads_Bsize));
+	  } else {
+	    ctx->macpads_Bsize = 64;
+	    ctx->macpads = gcry_malloc_secure (2*(ctx->macpads_Bsize));
+	  }
 	  if (! ctx->macpads)
 	    {
-	      md_close (hd);
 	      err = gpg_err_code_from_errno (errno);
+	      md_close (hd);
 	    }
 	}
     }
@@ -588,45 +599,54 @@ md_copy (gcry_md_hd_t ahd, gcry_md_hd_t *b_hd)
       b->debug = NULL;
       if (a->macpads)
 	{
-	  b->macpads = gcry_malloc_secure (128);
+	  b->macpads = gcry_malloc_secure (2*(a->macpads_Bsize));
 	  if (! b->macpads)
 	    {
-	      md_close (bhd);
 	      err = gpg_err_code_from_errno (errno);
+	      md_close (bhd);
 	    }
 	  else
-	    memcpy (b->macpads, a->macpads, 128);
+	    memcpy (b->macpads, a->macpads, (2*(a->macpads_Bsize)));
 	}
     }
 
   /* Copy the complete list of algorithms.  The copied list is
      reversed, but that doesn't matter. */
-  if (! err)
-    for (ar = a->list; ar; ar = ar->next)
-      {
-	if (a->secure)
-	  br = gcry_xmalloc_secure (sizeof *br
-				    + ar->digest->contextsize
-				    - sizeof(ar->context));
-	else
-	  br = gcry_xmalloc (sizeof *br
-			     + ar->digest->contextsize
-			     - sizeof (ar->context));
-	memcpy (br, ar,
-		sizeof (*br) + ar->digest->contextsize - sizeof (ar->context));
-	br->next = b->list;
-	b->list = br;
+  if (!err)
+    {
+      for (ar = a->list; ar; ar = ar->next)
+        {
+          if (a->secure)
+            br = gcry_malloc_secure (sizeof *br
+                                     + ar->digest->contextsize
+                                     - sizeof(ar->context));
+          else
+            br = gcry_malloc (sizeof *br
+                              + ar->digest->contextsize
+                              - sizeof (ar->context));
+          if (!br)
+            {
+	      err = gpg_err_code_from_errno (errno);
+              md_close (bhd);
+              break;
+            }
 
-	/* Add a reference to the module.  */
-	ath_mutex_lock (&digests_registered_lock);
-	_gcry_module_use (br->module);
-	ath_mutex_unlock (&digests_registered_lock);
-       }
+          memcpy (br, ar, (sizeof (*br) + ar->digest->contextsize
+                           - sizeof (ar->context)));
+          br->next = b->list;
+          b->list = br;
+          
+          /* Add a reference to the module.  */
+          ath_mutex_lock (&digests_registered_lock);
+          _gcry_module_use (br->module);
+          ath_mutex_unlock (&digests_registered_lock);
+        }
+    }
 
-  if (a->debug)
+  if (a->debug && !err)
     md_start_debug (bhd, "unknown");
 
-  if (! err)
+  if (!err)
     *b_hd = bhd;
 
   return err;
@@ -658,7 +678,7 @@ gcry_md_reset (gcry_md_hd_t a)
       (*r->digest->init) (&r->context.c);
     }
   if (a->ctx->macpads)
-    md_write (a, a->ctx->macpads, 64); /* inner pad */
+    md_write (a, a->ctx->macpads, a->ctx->macpads_Bsize); /* inner pad */
 }
 
 static void
@@ -682,7 +702,7 @@ md_close (gcry_md_hd_t a)
 
   if (a->ctx->macpads)
     {
-      wipememory (a->ctx->macpads, 128);
+      wipememory (a->ctx->macpads, 2*(a->ctx->macpads_Bsize));
       gcry_free(a->ctx->macpads);
     }
 
@@ -697,7 +717,7 @@ gcry_md_close (gcry_md_hd_t hd)
 }
 
 static void
-md_write (gcry_md_hd_t a, byte *inbuf, size_t inlen)
+md_write (gcry_md_hd_t a, const void *inbuf, size_t inlen)
 {
   GcryDigestEntry *r;
   
@@ -721,7 +741,7 @@ md_write (gcry_md_hd_t a, byte *inbuf, size_t inlen)
 void
 gcry_md_write (gcry_md_hd_t hd, const void *inbuf, size_t inlen)
 {
-  md_write (hd, (unsigned char *) inbuf, inlen);
+  md_write (hd, inbuf, inlen);
 }
 
 static void
@@ -751,7 +771,7 @@ md_final (gcry_md_hd_t a)
 
       if (err)
 	_gcry_fatal_error (err, NULL);
-      md_write (om, a->ctx->macpads+64, 64);
+      md_write (om, (a->ctx->macpads)+(a->ctx->macpads_Bsize), a->ctx->macpads_Bsize);
       md_write (om, p, dlen);
       md_final (om);
       /* Replace our digest with the mac (they have the same size). */
@@ -782,12 +802,12 @@ prepare_macpads( gcry_md_hd_t hd, const byte *key, size_t keylen)
       assert ( keylen <= 64 );
     }
 
-  memset ( hd->ctx->macpads, 0, 128 );
+  memset ( hd->ctx->macpads, 0, 2*(hd->ctx->macpads_Bsize) );
   ipad = hd->ctx->macpads;
-  opad = hd->ctx->macpads+64;
+  opad = (hd->ctx->macpads)+(hd->ctx->macpads_Bsize);
   memcpy ( ipad, key, keylen );
   memcpy ( opad, key, keylen );
-  for (i=0; i < 64; i++ ) 
+  for (i=0; i < (hd->ctx->macpads_Bsize); i++ ) 
     {
       ipad[i] ^= 0x36;
       opad[i] ^= 0x5c;
@@ -800,7 +820,6 @@ prepare_macpads( gcry_md_hd_t hd, const byte *key, size_t keylen)
 gcry_error_t
 gcry_md_ctl (gcry_md_hd_t hd, int cmd, void *buffer, size_t buflen)
 {
-  unsigned char *buf = (unsigned char *)buffer;
   gcry_err_code_t rc = 0;
   
   switch (cmd)
@@ -809,13 +828,13 @@ gcry_md_ctl (gcry_md_hd_t hd, int cmd, void *buffer, size_t buflen)
       md_final (hd);
       break;
     case GCRYCTL_SET_KEY:
-      rc = gcry_err_code (gcry_md_setkey (hd, buf, buflen));
+      rc = gcry_err_code (gcry_md_setkey (hd, buffer, buflen));
       break;
     case GCRYCTL_START_DUMP:
-      md_start_debug (hd, (char*)buf);
+      md_start_debug (hd, buffer);
       break;
     case GCRYCTL_STOP_DUMP:
-      md_stop_debug( hd );
+      md_stop_debug ( hd );
       break;
     default:
       rc = GPG_ERR_INV_OP;
@@ -840,6 +859,19 @@ gcry_md_setkey (gcry_md_hd_t hd, const void *key, size_t keylen)
   return gcry_error (rc);
 }
 
+/* The new debug interface.  If SUFFIX is a string it creates an debug
+   file for the context HD.  IF suffix is NULL, the file is closed and
+   debugging is stopped.  */
+void
+gcry_md_debug (gcry_md_hd_t hd, const char *suffix)
+{
+  if (suffix)
+    md_start_debug (hd, suffix);
+  else
+    md_stop_debug (hd);
+}
+
+
 
 /****************
  * if ALGO is null get the digest for the used algo (which should be only one)
@@ -853,7 +885,7 @@ md_read( gcry_md_hd_t a, int algo )
     {
       /* return the first algorithm */
       if (r && r->next)
-	log_debug("more than algorithm in md_read(0)\n");
+	log_debug ("more than one algorithm in md_read(0)\n");
       return r->digest->read( &r->context.c );
     }
   else
@@ -935,6 +967,11 @@ md_digest( gcry_md_hd_t a, int algo, byte *buffer, int buflen )
 gcry_err_code_t
 gcry_md_get (gcry_md_hd_t hd, int algo, byte *buffer, int buflen)
 {
+  (void)hd;
+  (void)algo;
+  (void)buffer;
+  (void)buflen;
+
   /*md_digest ... */
   return GPG_ERR_INTERNAL;
 }
@@ -1082,27 +1119,31 @@ gcry_md_algo_info (int algo, int what, void *buffer, size_t *nbytes)
       break;
 
     case GCRYCTL_GET_ASNOID:
-      {
-	const char unsigned *asn;
-	size_t asnlen;
-
-	asn = md_asn_oid (algo, &asnlen, NULL);
-	if (buffer && (*nbytes >= asnlen))
+      /* We need to check that the algo is available because
+         md_asn_oid would otherwise raise an assertion. */
+      err = check_digest_algo (algo);
+      if (!err)
+        {
+          const char unsigned *asn;
+          size_t asnlen;
+          
+          asn = md_asn_oid (algo, &asnlen, NULL);
+          if (buffer && (*nbytes >= asnlen))
 	  {
 	    memcpy (buffer, asn, asnlen);
 	    *nbytes = asnlen;
 	  }
-	else if ((! buffer) && nbytes)
-	  *nbytes = asnlen;
-	else
-	  {
-	    if (buffer)
-	      err = GPG_ERR_TOO_SHORT;
-	    else
-	      err = GPG_ERR_INV_ARG;
-	  }
-	break;
-      }
+          else if (!buffer && nbytes)
+            *nbytes = asnlen;
+          else
+            {
+              if (buffer)
+                err = GPG_ERR_TOO_SHORT;
+              else
+                err = GPG_ERR_INV_ARG;
+            }
+        }
+      break;
 
   default:
     err = GPG_ERR_INV_OP;
@@ -1113,31 +1154,34 @@ gcry_md_algo_info (int algo, int what, void *buffer, size_t *nbytes)
 
 
 static void
-md_start_debug( gcry_md_hd_t md, char *suffix )
+md_start_debug ( gcry_md_hd_t md, const char *suffix )
 {
   static int idx=0;
   char buf[50];
-
-  if( md->ctx->debug ) {
-    log_debug("Oops: md debug already started\n");
-    return;
-  }
+  
+  if ( md->ctx->debug )
+    {
+      log_debug("Oops: md debug already started\n");
+      return;
+    }
   idx++;
-  sprintf(buf, "dbgmd-%05d.%.10s", idx, suffix );
+  snprintf (buf, DIM(buf)-1, "dbgmd-%05d.%.10s", idx, suffix );
   md->ctx->debug = fopen(buf, "w");
-  if( !md->ctx->debug )
+  if ( !md->ctx->debug )
     log_debug("md debug: can't open %s\n", buf );
 }
 
 static void
 md_stop_debug( gcry_md_hd_t md )
 {
-  if( md->ctx->debug ) {
-    if( md->bufpos )
-      md_write( md, NULL, 0 );
-    fclose(md->ctx->debug);
-    md->ctx->debug = NULL;
-  }
+  if ( md->ctx->debug )
+    {
+      if ( md->bufpos )
+        md_write ( md, NULL, 0 );
+      fclose (md->ctx->debug);
+      md->ctx->debug = NULL;
+    }
+
 #ifdef HAVE_U64_TYPEDEF
   {  /* a kludge to pull in the __muldi3 for Solaris */
     volatile u32 a = (u32)(ulong)md;
@@ -1156,7 +1200,7 @@ md_stop_debug( gcry_md_hd_t md )
  *	Returns 1 when the handle works on secured memory
  *	otherwise 0 is returned.  There is no error return.
  *  GCRYCTL_IS_ALGO_ENABLED:
- *     Returns 1 if the algo is enanled for that handle.
+ *     Returns 1 if the algo is enabled for that handle.
  *     The algo must be passed as the address of an int.
  */
 gcry_error_t
