@@ -1,5 +1,5 @@
 /* random.h - random functions
- *	Copyright (C) 1998, 2002 Free Software Foundation, Inc.
+ *	Copyright (C) 1998, 2002, 2006 Free Software Foundation, Inc.
  *
  * This file is part of Libgcrypt.
  *
@@ -22,19 +22,34 @@
 
 #include "types.h"
 
-void _gcry_random_initialize (int);
+void _gcry_random_initialize (int full);
 void _gcry_register_random_progress (void (*cb)(void *,const char*,int,int,int),
                                      void *cb_data );
 void _gcry_random_dump_stats(void);
 void _gcry_secure_random_alloc(void);
-int  _gcry_quick_random_gen( int onoff );
+void _gcry_enable_quick_random_gen (void);
 int  _gcry_random_is_faked(void);
+void _gcry_set_random_daemon_socket (const char *socketname);
+int  _gcry_use_random_daemon (int onoff);
 void _gcry_set_random_seed_file (const char *name);
 void _gcry_update_random_seed_file (void);
 
 byte *_gcry_get_random_bits( size_t nbits, int level, int secure );
 void _gcry_fast_random_poll( void );
 
+/*-- rndegd.c --*/
+gpg_error_t _gcry_rndegd_set_socket_name (const char *name);
+
+/*-- random-daemon.c (only used from random.c) --*/
+#ifdef USE_RANDOM_DAEMON
+void _gcry_daemon_initialize_basics (void);
+int _gcry_daemon_randomize (const char *socketname,
+                            void *buffer, size_t length,
+                            enum gcry_random_level level);
+int _gcry_daemon_create_nonce (const char *socketname,
+                               void *buffer, size_t length);
+#endif /*USE_RANDOM_DAEMON*/
+
 #endif /*G10_RANDOM_H*/
 
 
