@@ -1,5 +1,6 @@
 /* mpiutil.ac  -  Utility functions for MPI
- * Copyright (C) 1998, 2000, 2001, 2002, 2003 Free Software Foundation, Inc.
+ * Copyright (C) 1998, 2000, 2001, 2002, 2003,
+ *               2007  Free Software Foundation, Inc.
  *
  * This file is part of Libgcrypt.
  *
@@ -14,8 +15,7 @@
  * GNU Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * License along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
@@ -27,13 +27,23 @@
 #include "g10lib.h"
 #include "mpi-internal.h"
 #include "memory.h"
+#include "mod-source-info.h"
+
+
+const char *
+_gcry_mpi_get_hw_config (void)
+{
+  return mod_source_info + 1;
+}
+
 
 /****************
  * Note:  It was a bad idea to use the number of limbs to allocate
  *	  because on a alpha the limbs are large but we normally need
- *	  integers of n bits - So we should chnage this to bits (or bytes).
+ *	  integers of n bits - So we should change this to bits (or bytes).
  *
- *	  But mpi_alloc is used in a lot of places :-)
+ *	  But mpi_alloc is used in a lot of places :-(.  New code
+ *	  should use mpi_new.
  */
 gcry_mpi_t
 _gcry_mpi_alloc( unsigned nlimbs )
@@ -115,7 +125,7 @@ _gcry_mpi_assign_limb_space( gcry_mpi_t a, mpi_ptr_t ap, unsigned int nlimbs )
 
 
 /****************
- * Resize the array of A to NLIMBS.  The additional space is cleared
+ * Resize the array of A to NLIMBS. The additional space is cleared
  * (set to 0).
  */
 void
@@ -132,6 +142,7 @@ _gcry_mpi_resize (gcry_mpi_t a, unsigned nlimbs)
       return; 
     }
 
+  /* Actually resize the limb space.  */
   if (a->d)
     {
       a->d = gcry_xrealloc (a->d, nlimbs * sizeof (mpi_limb_t));
@@ -231,7 +242,7 @@ gcry_mpi_get_opaque( gcry_mpi_t a, unsigned int *nbits )
  *	 but copy it transparently.
  */
 gcry_mpi_t
-_gcry_mpi_copy( gcry_mpi_t a )
+gcry_mpi_copy( gcry_mpi_t a )
 {
     int i;
     gcry_mpi_t b;
@@ -287,33 +298,69 @@ _gcry_mpi_alloc_like( gcry_mpi_t a )
 }
 
 
-void
-_gcry_mpi_set( gcry_mpi_t w, gcry_mpi_t u)
+gcry_mpi_t
+gcry_mpi_set( gcry_mpi_t w, gcry_mpi_t u)
 {
-    mpi_ptr_t wp, up;
-    mpi_size_t usize = u->nlimbs;
-    int usign = u->sign;
-
-    RESIZE_IF_NEEDED(w, usize);
-    wp = w->d;
-    up = u->d;
-    MPN_COPY( wp, up, usize );
-    w->nlimbs = usize;
-    w->flags = u->flags;
-    w->sign = usign;
+  mpi_ptr_t wp, up;
+  mpi_size_t usize = u->nlimbs;
+  int usign = u->sign;
+  
+  if (!w)
+    w = _gcry_mpi_alloc( mpi_get_nlimbs(u) );
+  RESIZE_IF_NEEDED(w, usize);
+  wp = w->d;
+  up = u->d;
+  MPN_COPY( wp, up, usize );
+  w->nlimbs = usize;
+  w->flags = u->flags;
+  w->sign = usign;
+  return w;
 }
 
 
-void
-_gcry_mpi_set_ui( gcry_mpi_t w, unsigned long u)
+gcry_mpi_t
+gcry_mpi_set_ui( gcry_mpi_t w, unsigned long u)
 {
-    RESIZE_IF_NEEDED(w, 1);
-    w->d[0] = u;
-    w->nlimbs = u? 1:0;
-    w->sign = 0;
-    w->flags = 0;
+  if (!w)
+    w = _gcry_mpi_alloc (1);
+  /* FIXME: If U is 0 we have no need to resize and thus possible
+     allocating the the limbs. */
+  RESIZE_IF_NEEDED(w, 1);
+  w->d[0] = u;
+  w->nlimbs = u? 1:0;
+  w->sign = 0;
+  w->flags = 0;
+  return w;
 }
 
+gcry_err_code_t
+_gcry_mpi_get_ui (gcry_mpi_t w, unsigned long *u)
+{
+  gcry_err_code_t err = GPG_ERR_NO_ERROR;
+  unsigned long x = 0;
+
+  if (w->nlimbs > 1)
+    err = GPG_ERR_TOO_LARGE;
+  else if (w->nlimbs == 1)
+    x = w->d[0];
+  else
+    x = 0;
+
+  if (! err)
+    *u = x;
+  
+  return err;
+}
+
+gcry_error_t
+gcry_mpi_get_ui (gcry_mpi_t w, unsigned long *u)
+{
+  gcry_err_code_t err = GPG_ERR_NO_ERROR;
+
+  err = _gcry_mpi_get_ui (w, u);
+  
+  return gcry_error (err);
+}
 
 gcry_mpi_t
 _gcry_mpi_alloc_set_ui( unsigned long u)
@@ -325,33 +372,28 @@ _gcry_mpi_alloc_set_ui( unsigned long u)
     return w;
 }
 
-
 void
-_gcry_mpi_swap( gcry_mpi_t a, gcry_mpi_t b)
+gcry_mpi_swap( gcry_mpi_t a, gcry_mpi_t b)
 {
     struct gcry_mpi tmp;
 
     tmp = *a; *a = *b; *b = tmp;
 }
 
-void
-gcry_mpi_swap( gcry_mpi_t a, gcry_mpi_t b)
-{
-  _gcry_mpi_swap (a, b);
-}
-
 
 gcry_mpi_t
 gcry_mpi_new( unsigned int nbits )
 {
-    return _gcry_mpi_alloc( (nbits+BITS_PER_MPI_LIMB-1) / BITS_PER_MPI_LIMB );
+    return _gcry_mpi_alloc ( (nbits+BITS_PER_MPI_LIMB-1)
+                             / BITS_PER_MPI_LIMB );
 }
 
 
 gcry_mpi_t
 gcry_mpi_snew( unsigned int nbits )
 {
-    return _gcry_mpi_alloc_secure( (nbits+BITS_PER_MPI_LIMB-1) / BITS_PER_MPI_LIMB );
+  return _gcry_mpi_alloc_secure ( (nbits+BITS_PER_MPI_LIMB-1)
+                                  / BITS_PER_MPI_LIMB );
 }
 
 void
@@ -360,31 +402,6 @@ gcry_mpi_release( gcry_mpi_t a )
     _gcry_mpi_free( a );
 }
 
-gcry_mpi_t
-gcry_mpi_copy( const gcry_mpi_t a )
-{
-    return _gcry_mpi_copy( (gcry_mpi_t)a );
-}
-
-gcry_mpi_t
-gcry_mpi_set( gcry_mpi_t w, const gcry_mpi_t u )
-{
-    if( !w )
-	w = _gcry_mpi_alloc( mpi_get_nlimbs(u) );
-    _gcry_mpi_set( w, (gcry_mpi_t)u );
-    return w;
-}
-
-gcry_mpi_t
-gcry_mpi_set_ui( gcry_mpi_t w, unsigned long u )
-{
-    if( !w )
-	w = _gcry_mpi_alloc(1);
-    _gcry_mpi_set_ui( w, u );
-    return w;
-}
-
-
 void
 gcry_mpi_randomize( gcry_mpi_t w,
 		    unsigned int nbits, enum gcry_random_level level )
@@ -421,22 +438,26 @@ gcry_mpi_set_flag( gcry_mpi_t a, enum gcry_mpi_flag flag )
 void
 gcry_mpi_clear_flag( gcry_mpi_t a, enum gcry_mpi_flag flag )
 {
-    switch( flag ) {
-      case GCRYMPI_FLAG_SECURE:
-      case GCRYMPI_FLAG_OPAQUE:
-      default: log_bug("invalid flag value\n");
+  (void)a; /* Not yet used. */
+
+  switch (flag)
+    {
+    case GCRYMPI_FLAG_SECURE:
+    case GCRYMPI_FLAG_OPAQUE:
+    default: log_bug("invalid flag value\n");
     }
 }
 
 int
 gcry_mpi_get_flag( gcry_mpi_t a, enum gcry_mpi_flag flag )
 {
-    switch( flag ) {
-      case GCRYMPI_FLAG_SECURE: return (a->flags & 1);
-      case GCRYMPI_FLAG_OPAQUE: return (a->flags & 4);
-      default: log_bug("invalid flag value\n");
+  switch (flag)
+    {
+    case GCRYMPI_FLAG_SECURE: return (a->flags & 1);
+    case GCRYMPI_FLAG_OPAQUE: return (a->flags & 4);
+    default: log_bug("invalid flag value\n");
     }
-    /*NOTREACHED*/
-    return 0;
+  /*NOTREACHED*/
+  return 0;
 }
 
