@@ -20,12 +20,12 @@
  */
 
 #include <config.h>
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <stdarg.h>
 #include <ctype.h>
-#include <assert.h>
 #include <limits.h>
 #include <errno.h>
 
@@ -35,22 +35,29 @@
 #include "secmem.h" /* our own secmem allocator */
 #include "ath.h"
 
+
+
 /****************
  * flag bits: 0 : general cipher debug
  *	      1 : general MPI debug
  */
 static unsigned int debug_flags;
 
+/* Controlled by global_init().  */
+static int any_init_done;
+
+/* Memory management. */
+
 static gcry_handler_alloc_t alloc_func;
 static gcry_handler_alloc_t alloc_secure_func;
 static gcry_handler_secure_check_t is_secure_func;
 static gcry_handler_realloc_t realloc_func;
 static gcry_handler_free_t free_func;
 static gcry_handler_no_mem_t outofcore_handler;
+static void *outofcore_handler_value;
+static int no_secure_memory;
 
-static void *outofcore_handler_value = NULL;
-static int no_secure_memory = 0;
-static int any_init_done;
+
 
 /* This is our handmade constructor.  It gets called by any function
    likely to be called at startup.  The suggested way for an
@@ -59,30 +66,51 @@ static int any_init_done;
 static void
 global_init (void)
 {
-  gcry_err_code_t err = GPG_ERR_NO_ERROR;
+  gcry_error_t err = 0;
 
   if (any_init_done)
     return;
   any_init_done = 1;
 
   err = ath_init ();
-  if (! err)
-    err = _gcry_cipher_init ();
-  if (! err)
-    err = _gcry_md_init ();
-  if (! err)
-    err = _gcry_pk_init ();
+  if (err)
+    goto fail;
+
+  /* Before we do any other initialization we need to test available
+     hardware features.  */
+  _gcry_detect_hw_features ();
+
+  err = _gcry_cipher_init ();
+  if (err)
+    goto fail;
+  err = _gcry_md_init ();
+  if (err)
+    goto fail;
+  err = _gcry_pk_init ();
+  if (err)
+    goto fail;
 #if 0
-  if (! err)
-    err = _gcry_ac_init ();
+  /* FIXME? */
+  err = _gcry_ac_init ();
+  if (err)
+    goto fail;
 #endif
 
-  if (err)
-    /* FIXME?  */
-    BUG ();
+  return;
+
+ fail:
+  /* FIXME: use `err'?  */
+  BUG ();
 }
 
+
 
+/* Version number parsing.  */
+
+/* This function parses the first portion of the version number S and
+   stores it in *NUMBER.  On sucess, this function returns a pointer
+   into S starting with the first character, which is not part of the
+   initial number portion; on failure, NULL is returned.  */
 static const char*
 parse_version_number( const char *s, int *number )
 {
@@ -98,6 +126,14 @@ parse_version_number( const char *s, int *number )
     return val < 0? NULL : s;
 }
 
+/* This function breaks up the complete string-representation of the
+   version number S, which is of the following struture: <major
+   number>.<minor number>.<micro number><patch level>.  The major,
+   minor and micro number components will be stored in *MAJOR, *MINOR
+   and *MICRO.
+
+   On success, the last component, the patch level, will be returned;
+   in failure, NULL will be returned.  */
 
 static const char *
 parse_version_string( const char *s, int *major, int *minor, int *micro )
@@ -116,12 +152,13 @@ parse_version_string( const char *s, int *major, int *minor, int *micro )
     return s; /* patchlevel */
 }
 
-/****************
- * Check that the the version of the library is at minimum the requested one
- * and return the version string; return NULL if the condition is not
- * satisfied.  If a NULL is passed to this function, no check is done,
- * but the version string is simply returned.
- */
+/* If REQ_VERSION is non-NULL, check that the version of the library
+   is at minimum the requested one.  Returns the string representation
+   of the library version if the condition is satisfied; return NULL
+   if the requested version is newer than that of the library.
+
+   If a NULL is passed to this function, no check is done, but the
+   string representation of the library is simply returned.  */
 const char *
 gcry_check_version( const char *req_version )
 {
@@ -130,18 +167,28 @@ gcry_check_version( const char *req_version )
     int rq_major, rq_minor, rq_micro;
     const char *my_plvl, *rq_plvl;
 
+    /* Initialize library.  */
     global_init ();
+
     if ( !req_version )
+        /* Caller wants our version number.  */
 	return ver;
 
+    /* Parse own version number.  */
     my_plvl = parse_version_string( ver, &my_major, &my_minor, &my_micro );
     if ( !my_plvl )
-	return NULL;  /* very strange our own version is bogus */
+        /* very strange our own version is bogus.  Shouldn't we use
+	   assert() here and bail out in case this happens?  -mo.  */
+	return NULL;
+
+  /* Parse requested version number.  */
     rq_plvl = parse_version_string( req_version, &rq_major, &rq_minor,
 								&rq_micro );
     if ( !rq_plvl )
-	return NULL;  /* req version string is invalid */
+        /* req version string is invalid, this can happen.  */
+	return NULL;
 
+    /* Compare version numbers.  */
     if ( my_major > rq_major
 	|| (my_major == rq_major && my_minor > rq_minor)
 	|| (my_major == rq_major && my_minor == rq_minor
@@ -151,17 +198,64 @@ gcry_check_version( const char *req_version )
 				 && strcmp( my_plvl, rq_plvl ) >= 0) ) {
 	return ver;
     }
+
     return NULL;
 }
 
+
+static void
+print_config ( int (*fnc)(FILE *fp, const char *format, ...), FILE *fp)
+{
+  unsigned int hwf;
+  struct {
+    unsigned int flag;
+    const char *desc;
+  } hwflist[] = {
+    { HWF_PADLOCK_RNG, "padlock-rng" },
+    { HWF_PADLOCK_AES, "padlock-aes" },
+    { HWF_PADLOCK_SHA, "padlock-sha" },
+    { 0, NULL} 
+  };
+  int i;
+
+  fnc (fp, "version:%s:\n", VERSION);
+  fnc (fp, "ciphers:%s:\n", LIBGCRYPT_CIPHERS);
+  fnc (fp, "pubkeys:%s:\n", LIBGCRYPT_PUBKEY_CIPHERS);
+  fnc (fp, "digests:%s:\n", LIBGCRYPT_DIGESTS);
+  fnc (fp, "rnd-mod:"
+#if USE_RNDEGD
+                "egd:"
+#endif
+#if USE_RNDLINUX
+                "linux:"
+#endif
+#if USE_RNDUNIX
+                "unix:"
+#endif
+#if USE_RNDW32
+                "w32:"
+#endif
+       "\n");
+  fnc (fp, "mpi-asm:%s:\n", _gcry_mpi_get_hw_config ());
+  hwf = _gcry_get_hw_features ();
+  fnc (fp, "hwflist:");
+  for (i=0; hwflist[i].desc; i++)
+  if ( (hwf & hwflist[i].flag) )
+    fnc (fp, "%s:", hwflist[i].desc);
+  fnc (fp, "\n");
+}
+
+
+
+
+/* Command dispatcher function, acting as general control
+   function.  */
 gcry_error_t
-gcry_control (enum gcry_ctl_cmds cmd, ...)
+_gcry_vcontrol (enum gcry_ctl_cmds cmd, va_list arg_ptr)
 {
-  gcry_err_code_t err = GPG_ERR_NO_ERROR;
   static int init_finished = 0;
-  va_list arg_ptr;
+  gcry_err_code_t err = 0;
   
-  va_start (arg_ptr, cmd);
   switch (cmd)
     {
     case GCRYCTL_ENABLE_M_GUARD:
@@ -169,21 +263,16 @@ gcry_control (enum gcry_ctl_cmds cmd, ...)
       break;
 
     case GCRYCTL_ENABLE_QUICK_RANDOM:
-      _gcry_quick_random_gen (1);
+      _gcry_enable_quick_random_gen ();
       break;
 
-    case 51:  /* Should be GCRYCTL_FAKED_RANDOM_P but we want to sneak
-                 that into the API for the sake of GnuPG 1.9 - there
-                 we check the error code and print a warning message
-                 if that call is not supported.  The literal number is
-                 used there as well. */
+    case GCRYCTL_FAKED_RANDOM_P:
       /* Return an error if the RNG is faked one (i.e. enabled by
          ENABLE_QUICK_RANDOM. */
       if (_gcry_random_is_faked ())
         err = GPG_ERR_GENERAL;
       break;
 
-
     case GCRYCTL_DUMP_RANDOM_STATS:
       _gcry_random_dump_stats ();
       break;
@@ -290,23 +379,71 @@ gcry_control (enum gcry_ctl_cmds cmd, ...)
 
     case GCRYCTL_SET_THREAD_CBS:
       err = ath_install (va_arg (arg_ptr, void *), any_init_done);
+      if (! err)
+	global_init ();
       break;
 
     case GCRYCTL_FAST_POLL:
       /* We need to do make sure that the random pool is really
-         initialized so that the poll fucntion is not a NOP. */
+         initialized so that the poll function is not a NOP. */
       _gcry_random_initialize (1);
       _gcry_fast_random_poll (); 
       break;
 
+    case GCRYCTL_SET_RNDEGD_SOCKET:
+#if USE_RNDEGD
+      err = _gcry_rndegd_set_socket_name (va_arg (arg_ptr, const char *));
+#else
+      err = gpg_error (GPG_ERR_NOT_SUPPORTED);
+#endif
+      break;
+
+    case GCRYCTL_SET_RANDOM_DAEMON_SOCKET:
+      _gcry_set_random_daemon_socket (va_arg (arg_ptr, const char *));
+      break;
+
+    case GCRYCTL_USE_RANDOM_DAEMON:
+      /* We need to do make sure that the random pool is really
+         initialized so that the poll function is not a NOP. */
+      _gcry_random_initialize (1);
+      _gcry_use_random_daemon (!! va_arg (arg_ptr, int));
+      break;
+      
+      /* This command dumps information pertaining to the
+         configuration of libgcrypt to the given stream.  It may be
+         used before the intialization has been finished but not
+         before a gcry_version_check. */
+    case GCRYCTL_PRINT_CONFIG:
+      {
+        FILE *fp = va_arg (arg_ptr, FILE *);
+        print_config (fp?fprintf:_gcry_log_info_with_dummy_fp, fp);
+      }
+      break;
+
     default:
       err = GPG_ERR_INV_OP;
     }
 
-  va_end(arg_ptr);
   return gcry_error (err);
 }
 
+
+/* Command dispatcher function, acting as general control
+   function.  */
+gcry_error_t
+gcry_control (enum gcry_ctl_cmds cmd, ...)
+{
+  gcry_error_t err;
+  va_list arg_ptr;
+  
+  va_start (arg_ptr, cmd);
+  err = _gcry_vcontrol (cmd, arg_ptr);
+  va_end(arg_ptr);
+  return err;
+}
+
+
+
 /* Return a pointer to a string containing a description of the error
    code in the error value ERR.  */
 const char *
@@ -402,8 +539,8 @@ gcry_set_outofcore_handler( int (*f)( void*, size_t, unsigned int ),
     outofcore_handler_value = value;
 }
 
-gcry_err_code_t
-_gcry_malloc (size_t n, unsigned int flags, void **mem)
+static gcry_err_code_t
+do_malloc (size_t n, unsigned int flags, void **mem)
 {
   gcry_err_code_t err = 0;
   void *m;
@@ -425,6 +562,8 @@ _gcry_malloc (size_t n, unsigned int flags, void **mem)
 
   if (!m)
     {
+      /* Make sure that ERRNO has been set in case a user supplied
+         memory handler didn't it correctly. */
       if (!errno)
         errno = ENOMEM;
       err = gpg_err_code_from_errno (errno);
@@ -440,7 +579,7 @@ gcry_malloc (size_t n)
 {
   void *mem = NULL;
 
-  _gcry_malloc (n, 0, &mem);
+  do_malloc (n, 0, &mem);
 
   return mem;
 }
@@ -450,7 +589,7 @@ gcry_malloc_secure (size_t n)
 {
   void *mem = NULL;
 
-  _gcry_malloc (n, GCRY_ALLOC_FLAG_SECURE, &mem);
+  do_malloc (n, GCRY_ALLOC_FLAG_SECURE, &mem);
 
   return mem;
 }
@@ -468,6 +607,8 @@ gcry_is_secure (const void *a)
 void
 _gcry_check_heap( const void *a )
 {
+  (void)a;
+  
     /* FIXME: implement this*/
 #if 0
     if( some_handler )
@@ -485,7 +626,7 @@ gcry_realloc (void *a, size_t n)
   if (realloc_func)
     p = realloc_func (a, n);
   else
-    p = _gcry_private_realloc (a, n);
+    p =  _gcry_private_realloc (a, n);
   if (!p && !errno)
     errno = ENOMEM;
   return p;
@@ -589,7 +730,7 @@ gcry_xrealloc( void *a, size_t n )
 
     while ( !(p = gcry_realloc( a, n )) ) {
 	if( !outofcore_handler
-	    || !outofcore_handler( outofcore_handler_value, n, 
+	    || !outofcore_handler( outofcore_handler_value, n,
                                    gcry_is_secure(a)? 3:2 ) ) {
 	    _gcry_fatal_error(gpg_err_code_from_errno (errno), NULL );
 	}
@@ -612,6 +753,7 @@ gcry_xmalloc_secure( size_t n )
     return p;
 }
 
+
 void *
 gcry_xcalloc( size_t n, size_t m )
 {
