@@ -1,6 +1,6 @@
 /* sexp.c  -  S-Expression handling
  * Copyright (C) 1999, 2000, 2001, 2002, 2003,
- *               2004 Free Software Foundation, Inc.
+ *               2004, 2006, 2007 Free Software Foundation, Inc.
  *
  * This file is part of Libgcrypt.
  *
@@ -26,7 +26,7 @@
 #include <string.h>
 #include <stdarg.h>
 #include <ctype.h>
-#include <assert.h>
+#include <errno.h>
 
 #define GCRYPT_NO_MPI_MACROS 1
 #include "g10lib.h"
@@ -54,14 +54,11 @@ struct gcry_sexp
 
 #define TOKEN_SPECIALS  "-./_:*+="
 
-
-
 static gcry_error_t
 sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	    const char *buffer, size_t length, int argflag,
 	    va_list arg_ptr, void **arg_list);
 
-
 /* Return true if P points to a byte containing a whitespace according
    to the S-expressions definition. */
 #undef whitespacep
@@ -177,16 +174,18 @@ normalize ( gcry_sexp_t list )
   if ( !list )
     return NULL;
   p = list->d;
-  if ( *p == ST_STOP ) {
-    /* this is "" */
-    gcry_sexp_release ( list );
-    return NULL;
-  }
-  if( *p == ST_OPEN && p[1] == ST_CLOSE ) {
-    /* this is "()" */
-    gcry_sexp_release ( list );
-    return NULL;
-  }
+  if ( *p == ST_STOP ) 
+    {
+      /* this is "" */
+      gcry_sexp_release ( list );
+      return NULL;
+    }
+  if ( *p == ST_OPEN && p[1] == ST_CLOSE )
+    {
+      /* this is "()" */
+      gcry_sexp_release ( list );
+      return NULL;
+    }
   
   return list;
 }
@@ -194,7 +193,7 @@ normalize ( gcry_sexp_t list )
 /* Create a new S-expression object by reading LENGTH bytes from
    BUFFER, assuming it is canonilized encoded or autodetected encoding
    when AUTODETECT is set to 1.  With FREEFNC not NULL, ownership of
-   the buffer is transferred to tyhe newle created object.  FREEFNC
+   the buffer is transferred to the newly created object.  FREEFNC
    should be the freefnc used to release BUFFER; there is no guarantee
    at which point this function is called; most likey you want to use
    free() or gcry_free(). 
@@ -243,7 +242,7 @@ gcry_sexp_create (gcry_sexp_t *retsexp, void *buffer, size_t length,
          have changed the internal represenation of S-expression to
          the canoncial format - which has the advantage of faster
          parsing - we will use this function as a closure in our
-         GCRYSEXP object and use the BUFFER directly */
+         GCRYSEXP object and use the BUFFER directly.  */
       freefnc (buffer);
     }
   return gcry_error (GPG_ERR_NO_ERROR);
@@ -308,6 +307,9 @@ gcry_sexp_release( gcry_sexp_t sexp )
 gcry_sexp_t
 gcry_sexp_cons( const gcry_sexp_t a, const gcry_sexp_t b )
 {
+  (void)a;
+  (void)b;
+
   /* NYI: Implementation should be quite easy with our new data
      representation */
   BUG ();
@@ -322,6 +324,8 @@ gcry_sexp_cons( const gcry_sexp_t a, const gcry_sexp_t b )
 gcry_sexp_t
 gcry_sexp_alist( const gcry_sexp_t *array )
 {
+  (void)array;
+
   /* NYI: Implementation should be quite easy with our new data
      representation. */
   BUG ();
@@ -334,6 +338,7 @@ gcry_sexp_alist( const gcry_sexp_t *array )
 gcry_sexp_t
 gcry_sexp_vlist( const gcry_sexp_t a, ... )
 {
+  (void)a;
   /* NYI: Implementation should be quite easy with our new data
      representation. */
   BUG ();
@@ -348,6 +353,8 @@ gcry_sexp_vlist( const gcry_sexp_t a, ... )
 gcry_sexp_t
 gcry_sexp_append( const gcry_sexp_t a, const gcry_sexp_t n )
 {
+  (void)a;
+  (void)n;
   /* NYI: Implementation should be quite easy with our new data
      representation. */
   BUG ();
@@ -357,6 +364,8 @@ gcry_sexp_append( const gcry_sexp_t a, const gcry_sexp_t n )
 gcry_sexp_t
 gcry_sexp_prepend( const gcry_sexp_t a, const gcry_sexp_t n )
 {
+  (void)a;
+  (void)n;
   /* NYI: Implementation should be quite easy with our new data
      representation. */
   BUG ();
@@ -372,61 +381,78 @@ gcry_sexp_prepend( const gcry_sexp_t a, const gcry_sexp_t n )
 gcry_sexp_t
 gcry_sexp_find_token( const gcry_sexp_t list, const char *tok, size_t toklen )
 {
-    const byte *p;
-    DATALEN n;
+  const byte *p;
+  DATALEN n;
+  
+  if ( !list )
+    return NULL;
 
-    if ( !list )
-	return NULL;
+  if ( !toklen )
+    toklen = strlen(tok);
 
-    if( !toklen )
-	toklen = strlen(tok);
-    p = list->d;
-    while ( *p != ST_STOP ) {
-	if ( *p == ST_OPEN && p[1] == ST_DATA ) {
-	    const byte *head = p;
+  p = list->d;
+  while ( *p != ST_STOP )
+    {
+      if ( *p == ST_OPEN && p[1] == ST_DATA ) 
+        {
+          const byte *head = p;
 
-	    p += 2;
-	    memcpy ( &n, p, sizeof n ); p += sizeof n;
-	    if ( n == toklen && !memcmp( p, tok, toklen ) ) { /* found it */
-		gcry_sexp_t newlist;
-		byte *d;
-		int level = 1;
+          p += 2;
+          memcpy ( &n, p, sizeof n );
+          p += sizeof n;
+          if ( n == toklen && !memcmp( p, tok, toklen ) )
+            { /* found it */
+              gcry_sexp_t newlist;
+              byte *d;
+              int level = 1;
 
-		/* look for the end of the list */
-		for ( p += n; level; p++ ) {
-		    if ( *p == ST_DATA ) {
+              /* Look for the end of the list.  */
+              for ( p += n; level; p++ ) 
+                {
+                  if ( *p == ST_DATA )
+                    {
 			memcpy ( &n, ++p, sizeof n );
 			p += sizeof n + n;
-			p--; /* compensate for later increment */
+			p--; /* Compensate for later increment. */
 		    }
-		    else if ( *p == ST_OPEN ) {
-			level++;
+                  else if ( *p == ST_OPEN ) 
+                    {
+                      level++;
 		    }
-		    else if ( *p == ST_CLOSE ) {
-			level--;
+                  else if ( *p == ST_CLOSE ) 
+                    {
+                      level--;
 		    }
-		    else if ( *p == ST_STOP ) {
-			BUG ();
+                  else if ( *p == ST_STOP ) 
+                    {
+                      BUG ();
 		    }
-		} while ( level );
-		n = p - head;
+		}
+              n = p - head;
 
-		newlist = gcry_xmalloc ( sizeof *newlist + n );
-		d = newlist->d;
-		memcpy ( d, head, n ); d += n;
-		*d++ = ST_STOP;
-		return normalize ( newlist );
+              newlist = gcry_malloc ( sizeof *newlist + n );
+              if (!newlist)
+                {
+                  /* No way to return an error code, so we can only
+                     return Not Found. */
+                  return NULL;
+                }
+              d = newlist->d;
+              memcpy ( d, head, n ); d += n;
+              *d++ = ST_STOP;
+              return normalize ( newlist );
 	    }
-	    p += n;
+          p += n;
 	}
-	else if ( *p == ST_DATA ) {
-	    memcpy ( &n, ++p, sizeof n ); p += sizeof n;
-	    p += n;
+      else if ( *p == ST_DATA )
+        {
+          memcpy ( &n, ++p, sizeof n ); p += sizeof n;
+          p += n;
 	}
-	else
-	    p++;
+      else
+        p++;
     }
-    return NULL;
+  return NULL;
 }
 
 /****************
@@ -467,9 +493,8 @@ gcry_sexp_length( const gcry_sexp_t list )
 
 
 
-/****************
- * Extract the CAR of the given list
- */
+/* Extract the CAR of the given list.  May return NULL for bad lists
+   or memory failure.  */
 gcry_sexp_t
 gcry_sexp_nth( const gcry_sexp_t list, int number )
 {
@@ -508,7 +533,9 @@ gcry_sexp_nth( const gcry_sexp_t list, int number )
 
     if ( *p == ST_DATA ) {
 	memcpy ( &n, p, sizeof n ); p += sizeof n;
-	newlist = gcry_xmalloc ( sizeof *newlist + n + 1 );
+	newlist = gcry_malloc ( sizeof *newlist + n + 1 );
+        if (!newlist)
+          return NULL;
 	d = newlist->d;
 	memcpy ( d, p, n ); d += n;
 	*d++ = ST_STOP;
@@ -536,7 +563,9 @@ gcry_sexp_nth( const gcry_sexp_t list, int number )
 	} while ( level );
 	n = p + 1 - head;
 
-	newlist = gcry_xmalloc ( sizeof *newlist + n );
+	newlist = gcry_malloc ( sizeof *newlist + n );
+        if (!newlist)
+          return NULL;
 	d = newlist->d;
 	memcpy ( d, head, n ); d += n;
 	*d++ = ST_STOP;
@@ -553,116 +582,116 @@ gcry_sexp_car( const gcry_sexp_t list )
     return gcry_sexp_nth ( list, 0 );
 }
 
-/****************
- * Get data from the car.  The returned value is valid as long as the list
- * is not modified.
- */
-const char *
-gcry_sexp_nth_data( const gcry_sexp_t list, int number, size_t *datalen )
+
+/* Helper to get data from the car.  The returned value is valid as
+   long as the list is not modified. */
+static const char *
+sexp_nth_data (const gcry_sexp_t list, int number, size_t *datalen)
 {
-    const unsigned char *p;
-    DATALEN n;
-    int level = 0;
-
-    *datalen = 0;
-    if ( !list ) {
-	return NULL;
-    }
-    p = list->d;
-    if ( *p == ST_OPEN )
-	p++;	     /* yep, a list */
-    else if (number )
-	return NULL; /* not a list but an n > 0 element requested */
-
-    /* skip n elements */
-    while ( number > 0 ) {
-	if ( *p == ST_DATA ) {
-	    memcpy ( &n, ++p, sizeof n );
-	    p += sizeof n + n;
-	    p--;
-	    if ( !level )
-		number--;
-	}
-	else if ( *p == ST_OPEN ) {
-	    level++;
-	}
-	else if ( *p == ST_CLOSE ) {
-	    level--;
-	    if ( !level )
-		number--;
-	}
-	else if ( *p == ST_STOP ) {
-	    return NULL;
-	}
-	p++;
-    }
-
-
-    if ( *p == ST_DATA ) {
-	memcpy ( &n, ++p, sizeof n );
-	*datalen = n;
-	return (const char*)p + sizeof n;
-    }
-
+  const byte *p;
+  DATALEN n;
+  int level = 0;
+  
+  *datalen = 0;
+  if ( !list ) 
     return NULL;
+
+  p = list->d;
+  if ( *p == ST_OPEN )
+    p++;	     /* Yep, a list. */
+  else if (number)
+    return NULL;     /* Not a list but N > 0 requested. */
+
+  /* Skip over N elements. */
+  while ( number > 0 ) 
+    {
+      if ( *p == ST_DATA ) 
+        {
+          memcpy ( &n, ++p, sizeof n );
+          p += sizeof n + n;
+          p--;
+          if ( !level )
+            number--;
+	}
+      else if ( *p == ST_OPEN ) 
+        {
+          level++;
+	}
+      else if ( *p == ST_CLOSE ) 
+        {
+          level--;
+          if ( !level )
+            number--;
+	}
+      else if ( *p == ST_STOP ) 
+        {
+          return NULL;
+	}
+      p++;
+    }
+
+  /* If this is data, return it.  */
+  if ( *p == ST_DATA )
+    {
+      memcpy ( &n, ++p, sizeof n );
+      *datalen = n;
+      return (const char*)p + sizeof n;
+    }
+  
+  return NULL;
 }
 
-/****************
+
+/* Get data from the car.  The returned value is valid as long as the
+   list is not modified.  */
+const char *
+gcry_sexp_nth_data (const gcry_sexp_t list, int number, size_t *datalen )
+{
+  return sexp_nth_data (list, number, datalen);
+}
+
+
+/* Get a string from the car.  The returned value is a malloced string
+   and needs to be freed by the caller.  */
+char *
+gcry_sexp_nth_string (const gcry_sexp_t list, int number)
+{
+  const char *s;
+  size_t n;
+  char *buf;
+
+  s = sexp_nth_data (list, number, &n);
+  if (!s || n < 1 || (n+1) < 1)
+    return NULL;
+  buf = gcry_malloc (n+1);
+  if (!buf)
+    return NULL;
+  memcpy (buf, s, n);
+  buf[n] = 0;
+  return buf;
+}
+
+/*
  * Get a MPI from the car
  */
 gcry_mpi_t
 gcry_sexp_nth_mpi( gcry_sexp_t list, int number, int mpifmt )
 {
-    const byte *p;
-    DATALEN n;
-    int level = 0;
+  const char *s;
+  size_t n;
+  gcry_mpi_t a;
 
-    if ( !list )
-	return NULL;
-    if ( !mpifmt )
-	mpifmt = GCRYMPI_FMT_STD;
-
-    p = list->d;
-    if ( *p == ST_OPEN )
-	p++;	     /* yep, a list */
-    else if (number )
-	return NULL; /* not a list but an n > 0 element requested */
-
-    /* skip n elements */
-    while ( number > 0 ) {
-	if ( *p == ST_DATA ) {
-	    memcpy ( &n, ++p, sizeof n );
-	    p += sizeof n + n;
-	    p--;
-	    if ( !level )
-		number--;
-	}
-	else if ( *p == ST_OPEN ) {
-	    level++;
-	}
-	else if ( *p == ST_CLOSE ) {
-	    level--;
-	    if ( !level )
-		number--;
-	}
-	else if ( *p == ST_STOP ) {
-	    return NULL;
-	}
-	p++;
-    }
-
-    if ( *p == ST_DATA ) {
-	gcry_mpi_t a;
-	size_t nbytes;
-
-	memcpy ( &n, ++p, sizeof n );
-	p += sizeof n;
-	nbytes = n;
-	if( !gcry_mpi_scan( &a, mpifmt, p, n, &nbytes ) )
-	    return a;
-    }
+  if ( !mpifmt )
+    mpifmt = GCRYMPI_FMT_STD;
 
+  s = sexp_nth_data (list, number, &n);
+  if (!s)
     return NULL;
+
+  if ( gcry_mpi_scan ( &a, mpifmt, s, n, NULL ) )
+    return NULL;
+
+  return a;
 }
 
 
@@ -728,7 +757,9 @@ gcry_sexp_cdr( const gcry_sexp_t list )
     } while ( level );
     n = p - head;
 
-    newlist = gcry_xmalloc ( sizeof *newlist + n + 2 );
+    newlist = gcry_malloc ( sizeof *newlist + n + 2 );
+    if (!newlist)
+      return NULL;
     d = newlist->d;
     *d++ = ST_OPEN;
     memcpy ( d, head, n ); d += n;
@@ -752,7 +783,7 @@ gcry_sexp_cadr ( const gcry_sexp_t list )
 
 
 static int
-hextobyte( const unsigned char *s )
+hextobyte( const byte *s )
 {
     int c=0;
 
@@ -774,28 +805,35 @@ hextobyte( const unsigned char *s )
     return c;
 }
 
-struct make_space_ctx 
-{
-  gcry_sexp_t sexp;
-  size_t allocated;
-  unsigned char *pos;
+struct make_space_ctx {
+    gcry_sexp_t sexp;
+    size_t allocated;
+    byte *pos;
 };
 
-static void
+static gpg_err_code_t
 make_space ( struct make_space_ctx *c, size_t n )
 {
-    size_t used = c->pos - c->sexp->d;
-
-    if ( used + n + sizeof(DATALEN) + 1 >= c->allocated ) {
-	gcry_sexp_t newsexp;
-	byte *newhead;
-
-	c->allocated += 2*(n+sizeof(DATALEN)+1);
-	newsexp = gcry_xrealloc ( c->sexp, sizeof *newsexp + c->allocated - 1 );
-	newhead = newsexp->d;
-	c->pos = newhead + used;
-	c->sexp = newsexp;
+  size_t used = c->pos - c->sexp->d;
+  
+  if ( used + n + sizeof(DATALEN) + 1 >= c->allocated )
+    {
+      gcry_sexp_t newsexp;
+      byte *newhead;
+      size_t newsize;
+      
+      newsize = c->allocated + 2*(n+sizeof(DATALEN)+1);
+      if (newsize <= c->allocated)
+        return GPG_ERR_TOO_LARGE;
+      newsexp = gcry_realloc ( c->sexp, sizeof *newsexp + newsize - 1);
+      if (!newsexp)
+        return gpg_err_code_from_errno (errno);
+      c->allocated = newsize;
+      newhead = newsexp->d;
+      c->pos = newhead + used;
+      c->sexp = newsexp;
     }
+  return 0;
 }
 
 
@@ -803,11 +841,10 @@ make_space ( struct make_space_ctx *c, size_t n )
    quotes are must already be removed from STRING.  We assume that the
    quoted string is syntacillay correct.  */
 static size_t
-unquote_string (const char *string_arg, size_t length, unsigned char *buf)
+unquote_string (const char *string, size_t length, unsigned char *buf)
 {
-  const unsigned char *string = (const unsigned char*)string_arg;
   int esc = 0;
-  const unsigned char *s = string;
+  const unsigned char *s = (const unsigned char*)string;
   unsigned char *d = buf;
   size_t n = length;
 
@@ -898,7 +935,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	    const char *buffer, size_t length, int argflag,
 	    va_list arg_ptr, void **arg_list)
 {
-  gcry_err_code_t err = GPG_ERR_NO_ERROR;
+  gcry_err_code_t err = 0;
   static const char tokenchars[] =
     "abcdefghijklmnopqrstuvwxyz"
     "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
@@ -920,8 +957,6 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
   int arg_counter = 0;
   int level = 0;
 
-  /* FIXME: invent better error codes (?).  */
-
   if (!erroff)
     erroff = &dummy_erroff;
 
@@ -938,21 +973,41 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
     }                                                    \
   while (0)
 
-#define MAKE_SPACE(n)  do { make_space ( &c, (n) ); } while (0)
+  /* The MAKE_SPACE macro is used before each store operation to
+     ensure that the buffer is large enough.  It requires a global
+     context named C and jumps out to the label LEAVE on error! It
+     also sets ERROFF using the variables BUFFER and P.  */
+#define MAKE_SPACE(n)  do {                                                \
+                            gpg_err_code_t _ms_err = make_space (&c, (n)); \
+                            if (_ms_err)                                   \
+                              {                                            \
+                                err = _ms_err;                             \
+                                *erroff = p - buffer;                      \
+                                goto leave;                                \
+                              }                                            \
+                       } while (0)
+
+  /* The STORE_LEN macro is used to store the length N at buffer P. */
 #define STORE_LEN(p,n) do {						   \
 			    DATALEN ashort = (n);			   \
 			    memcpy ( (p), &ashort, sizeof(ashort) );	   \
 			    (p) += sizeof (ashort);			   \
 			} while (0)
 
-  /* We assume that the internal representation takes less memory
-   * than the provided one.  However, we add space for one extra datalen
-   * so that the code which does the ST_CLOSE can use MAKE_SPACE */
+  /* We assume that the internal representation takes less memory than
+     the provided one.  However, we add space for one extra datalen so
+     that the code which does the ST_CLOSE can use MAKE_SPACE */
   c.allocated = length + sizeof(DATALEN);
   if (buffer && length && gcry_is_secure (buffer))
-    c.sexp = gcry_xmalloc_secure (sizeof *c.sexp + c.allocated - 1);
+    c.sexp = gcry_malloc_secure (sizeof *c.sexp + c.allocated - 1);
   else
-    c.sexp = gcry_xmalloc (sizeof *c.sexp + c.allocated - 1);
+    c.sexp = gcry_malloc (sizeof *c.sexp + c.allocated - 1);
+  if (!c.sexp)
+    {
+      err = gpg_err_code_from_errno (errno);
+      *erroff = 0;
+      goto leave;
+    }
   c.pos = c.sexp->d;
 
   for (p = buffer, n = length; n; p++, n--)
@@ -993,6 +1048,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 		      *erroff = p - buffer;
 		      /* Invalid octal value.  */
 		      err = GPG_ERR_SEXP_BAD_QUOTATION;
+                      goto leave;
 		    }
 		  p += 2;
 		  n -= 2;
@@ -1005,6 +1061,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 		      *erroff = p - buffer;
 		      /* Invalid hex value.  */
 		      err = GPG_ERR_SEXP_BAD_QUOTATION;
+                      goto leave;
 		    }
 		  p += 2;
 		  n -= 2;
@@ -1035,6 +1092,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 		  *erroff = p - buffer;
 		  /* Invalid quoted string escape.  */
 		  err = GPG_ERR_SEXP_BAD_QUOTATION;
+                  goto leave;
 		}
 	    }
 	  else if (*p == '\\')
@@ -1067,6 +1125,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 		{
 		  *erroff = p - buffer;
 		  err = GPG_ERR_SEXP_ODD_HEX_NUMBERS;
+                  goto leave;
 		}
 
 	      datalen = hexcount / 2;
@@ -1086,6 +1145,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	    {
 	      *erroff = p - buffer;
 	      err = GPG_ERR_SEXP_BAD_HEX_CHAR;
+              goto leave;
 	    }
 	}
       else if (base64)
@@ -1106,6 +1166,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 		  *erroff = p - buffer;
 		  /* Buffer too short.  */
 		  err = GPG_ERR_SEXP_STRING_TOO_LONG;
+                  goto leave;
 		}
 	      /* Make a new list entry.  */
 	      MAKE_SPACE (datalen);
@@ -1137,6 +1198,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	    {
 	      *erroff = p - buffer;
 	      err = GPG_ERR_SEXP_INV_LEN_SPEC;
+              goto leave;
 	    }
 	}
       else if (percent)
@@ -1153,15 +1215,20 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 		BUG ();
 
 	      MAKE_SPACE (nm);
-	      if ((!gcry_is_secure (c.sexp->d))
-		  &&  gcry_mpi_get_flag ( m, GCRYMPI_FLAG_SECURE))
+	      if (!gcry_is_secure (c.sexp->d)
+		  && gcry_mpi_get_flag ( m, GCRYMPI_FLAG_SECURE))
 		{
 		  /* We have to switch to secure allocation.  */
 		  gcry_sexp_t newsexp;
 		  byte *newhead;
 
-		  newsexp = gcry_xmalloc_secure (sizeof *newsexp
-						 + c.allocated - 1);
+		  newsexp = gcry_malloc_secure (sizeof *newsexp
+                                                + c.allocated - 1);
+                  if (!newsexp)
+                    {
+                      err = gpg_err_code_from_errno (errno);
+                      goto leave;
+                    }
 		  newhead = newsexp->d;
 		  memcpy (newhead, c.sexp->d, (c.pos - c.sexp->d));
 		  c.pos = newhead + (c.pos - c.sexp->d);
@@ -1208,8 +1275,13 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 		  gcry_sexp_t newsexp;
 		  byte *newhead;
 
-		  newsexp = gcry_xmalloc_secure (sizeof *newsexp
- 						 + c.allocated - 1);
+		  newsexp = gcry_malloc_secure (sizeof *newsexp
+                                                + c.allocated - 1);
+                  if (!newsexp)
+                    {
+                      err = gpg_err_code_from_errno (errno);
+                      goto leave;
+                    }
 		  newhead = newsexp->d;
 		  memcpy (newhead, c.sexp->d, (c.pos - c.sexp->d));
 		  c.pos = newhead + (c.pos - c.sexp->d);
@@ -1243,6 +1315,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	      *erroff = p - buffer;
 	      /* Invalid format specifier.  */
 	      err = GPG_ERR_SEXP_INV_LEN_SPEC;
+              goto leave;
 	    }
 	  percent = NULL;
 	}
@@ -1253,6 +1326,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	      *erroff = p - buffer;
 	      /* Open display hint.  */
 	      err = GPG_ERR_SEXP_UNMATCHED_DH;
+              goto leave;
 	    }
 	  MAKE_SPACE (0);
 	  *c.pos++ = ST_OPEN;
@@ -1266,6 +1340,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	      *erroff = p - buffer;
 	      /* Open display hint.  */
 	      err = GPG_ERR_SEXP_UNMATCHED_DH;
+              goto leave;
 	    }
 	  MAKE_SPACE (0);
 	  *c.pos++ = ST_CLOSE;
@@ -1290,6 +1365,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	      *erroff = p - buffer;
 	      /* Open display hint.  */
 	      err = GPG_ERR_SEXP_NESTED_DH;
+              goto leave;
 	    }
 	  disphint = p;
 	}
@@ -1300,6 +1376,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	      *erroff = p - buffer;
 	      /* Open display hint.  */
 	      err = GPG_ERR_SEXP_UNMATCHED_DH;
+              goto leave;
 	    }
 	  disphint = NULL;
 	}
@@ -1310,6 +1387,7 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	      /* A length may not begin with zero.  */
 	      *erroff = p - buffer;
 	      err = GPG_ERR_SEXP_ZERO_PREFIX;
+              goto leave;
 	    }
 	  digptr = p;
 	}
@@ -1325,12 +1403,14 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	     need to save it.  Great.  */
 	  *erroff = p - buffer;
 	  err = GPG_ERR_SEXP_UNEXPECTED_PUNC;
+          goto leave;
 	}
       else if (strchr ("&\\", *p))
 	{
 	  /* Reserved punctuation.  */
 	  *erroff = p - buffer;
 	  err = GPG_ERR_SEXP_UNEXPECTED_PUNC;
+          goto leave;
 	}
       else if (argflag && (*p == '%'))
 	percent = p;
@@ -1339,14 +1419,16 @@ sexp_sscan (gcry_sexp_t *retsexp, size_t *erroff,
 	  /* Bad or unavailable.  */
 	  *erroff = p - buffer;
 	  err = GPG_ERR_SEXP_BAD_CHARACTER;
+          goto leave;
 	}
     }
   MAKE_SPACE (0);
   *c.pos++ = ST_STOP;
 
-  if (level)
+  if (level && !err)
     err = GPG_ERR_SEXP_UNMATCHED_PAREN;
 
+ leave:
   if (err)
     {
       /* Error -> deallocate.  */
@@ -1382,6 +1464,15 @@ gcry_sexp_build (gcry_sexp_t *retsexp, size_t *erroff, const char *format, ...)
   return rc;
 }
 
+
+gcry_error_t
+_gcry_sexp_vbuild (gcry_sexp_t *retsexp, size_t *erroff, 
+                   const char *format, va_list arg_ptr)
+{
+  return sexp_sscan (retsexp, erroff, format, strlen(format), 1,
+                     arg_ptr, NULL);
+}
+
 /* Like gcry_sexp_build, but uses an array instead of variable
    function arguments.  */
 gcry_error_t
@@ -1448,7 +1539,7 @@ suitable_encoding (const unsigned char *buffer, size_t length)
 
 
 static int
-convert_to_hex (const unsigned char *src, size_t len, unsigned char *dest)
+convert_to_hex (const unsigned char *src, size_t len, char *dest)
 {
   int i;
 
@@ -1456,18 +1547,18 @@ convert_to_hex (const unsigned char *src, size_t len, unsigned char *dest)
     {
       *dest++ = '#';
       for (i=0; i < len; i++, dest += 2 )
-        sprintf ((char*)dest, "%02X", src[i]);
+        sprintf (dest, "%02X", src[i]);
       *dest++ = '#';
     }
   return len*2+2;
 }
 
 static int
-convert_to_string (const unsigned char *s, size_t len, unsigned char *dest)
+convert_to_string (const unsigned char *s, size_t len, char *dest)
 {
   if (dest)
     {
-      unsigned char *p = dest;
+      char *p = dest;
       *p++ = '\"';
       for (; len; len--, s++ )
         {
@@ -1485,7 +1576,7 @@ convert_to_string (const unsigned char *s, size_t len, unsigned char *dest)
             default: 
               if ( (*s < 0x20 || (*s >= 0x7f && *s <= 0xa0)))
                 {
-                  sprintf ((char*)p, "\\x%02x", *s); 
+                  sprintf (p, "\\x%02x", *s); 
                   p += 4;
                 }
               else
@@ -1525,7 +1616,7 @@ convert_to_string (const unsigned char *s, size_t len, unsigned char *dest)
 
 
 static int
-convert_to_token (const unsigned char *src, size_t len, unsigned char *dest)
+convert_to_token (const unsigned char *src, size_t len, char *dest)
 {
   if (dest)
     memcpy (dest, src, len);
@@ -1541,7 +1632,7 @@ convert_to_token (const unsigned char *src, size_t len, unsigned char *dest)
  */
 size_t
 gcry_sexp_sprint (const gcry_sexp_t list, int mode,
-                  void *buffer, size_t maxlength)
+                  void *buffer, size_t maxlength )
 {
   static unsigned char empty[3] = { ST_OPEN, ST_CLOSE, ST_STOP };
   const unsigned char *s;
@@ -1626,9 +1717,9 @@ gcry_sexp_sprint (const gcry_sexp_t list, int mode,
                     return 0;
                   switch (type)
                     {
-                    case 1: convert_to_string (s, n, (unsigned char*)d); break;
-                    case 2: convert_to_token (s, n, (unsigned char*)d); break;
-                    default: convert_to_hex (s, n, (unsigned char*)d); break;
+                    case 1: convert_to_string (s, n, d); break;
+                    case 2: convert_to_token (s, n, d); break;
+                    default: convert_to_hex (s, n, d); break;
                     }
                   d += nn;
                 }
@@ -1695,7 +1786,7 @@ gcry_sexp_canon_len (const unsigned char *buffer, size_t length,
                      size_t *erroff, gcry_error_t *errcode)
 {
   const unsigned char *p;
-  const char *disphint=NULL;
+  const unsigned char *disphint = NULL;
   unsigned int datalen = 0;
   size_t dummy_erroff;
   gcry_error_t dummy_errcode;
@@ -1784,11 +1875,11 @@ gcry_sexp_canon_len (const unsigned char *buffer, size_t length,
               *errcode = gcry_error (GPG_ERR_SEXP_NESTED_DH);
               return 0;
             }
-          disphint = (const char*)p;
+          disphint = p;
 	}
       else if (*p == ']')
         {
-          if( !disphint ) 
+          if ( !disphint ) 
             {
               *erroff = count;
               *errcode = gcry_error (GPG_ERR_SEXP_UNMATCHED_DH);
